import argparse
import tempfile

from ming.orm import session
from pylons import tmpl_context as c
from pylons import app_globals as g
from tg import (
        config,
        expose,
        flash,
        redirect,
        )
from tg.decorators import (
        with_trailing_slash,
        without_trailing_slash,
        )

from allura.lib.decorators import require_post
from allura.lib import helpers as h
from allura.model import (
        AuditLog,
        OAuthConsumerToken,
        OAuthRequestToken,
        OAuthAccessToken
)

from forgeimporters.base import (
    ToolImporter,
    ToolImportForm,
    ToolImportController,
)
from forgeimporters.trac import TracURLValidator

from tracwikiimporter.scripts.wiki_from_trac.extractors import WikiExporter
from tracwikiimporter.scripts.wiki_from_trac.loaders import load_data
from tracwikiimporter.scripts.wiki_from_trac.wiki_from_trac import WikiFromTrac


class TracWikiImportForm(ToolImportForm):
    trac_url = TracURLValidator()


class TracWikiImportController(ToolImportController):
    import_form = TracWikiImportForm

    @with_trailing_slash
    @expose('jinja:tracwikiimporter:templates/index.html')
    def index(self, **kw):
        return dict(importer=self.importer,
                target_app=self.target_app)

    @without_trailing_slash
    @expose()
    @require_post()
    def create(self, trac_url, mount_point, mount_label, **kw):
        if self.importer.enforce_limit(c.project):
            self.importer.post(
                    mount_point=mount_point,
                    mount_label=mount_label,
                    trac_url=trac_url,
                    )
            flash('Wiki import has begun. Your new wiki will be available '
                    'when the import is complete.')
        else:
            flash('There are too many imports pending at this time.  Please wait and try again.', 'error')
        redirect(c.project.url() + 'admin/')


class TracWikiImporter(ToolImporter):
    target_app_ep_names = 'wiki'
    source = 'Trac'
    controller = TracWikiImportController
    tool_label = 'Wiki'
    tool_description = 'Import your wiki from Trac.  Note: wiki content is imported, but not revision history or attachments.'

    def import_tool(self, project, user, project_name=None, mount_point=None,
            mount_label=None, trac_url=None, **kw):
        """ Import Trac wiki into a new Allura Wiki tool.

        """
        mount_point = mount_point or 'wiki'
        app = project.install_app(
                'Wiki',
                mount_point=mount_point,
                mount_label=mount_label or 'Wiki',
                import_id={
                        'source': self.source,
                        'trac_url': trac_url,
                    },
            )
        session(app.config).flush(app.config)
        try:
            options = argparse.Namespace()
            options.token = self.get_token(user)
            options.wiki = mount_point
            options.base_url = config['base_url'].replace('http://', 'https://')
            options.verbose = False
            options.converter = 'html2text'
            options.import_opts = []
            options.user_map_file = None
            options.neighborhood, options.project = project.url().strip('/').split('/',1)
            with tempfile.NamedTemporaryFile() as f:
                WikiExporter(trac_url, options).export(f)
                f.flush()
                load_data(f.name, WikiFromTrac.parser(), options)
            AuditLog.log('import tool %s from %s' %
                    (app.config.options.mount_point, trac_url),
                    project=project, user=user, url=app.url)
            g.post_event('project_updated')
            return app
        except Exception as e:
            h.make_app_admin_only(app)
            raise

    def get_token(self, user):
        consumer_token = OAuthConsumerToken.upsert(self.classname, user)
        request_token_params = dict(
            consumer_token_id=consumer_token._id,
            user_id=user._id,
            callback='manual')
        request_token = OAuthRequestToken.query.get(**request_token_params)
        if request_token is None:
            request_token_params.update(dict(validation_pin=h.nonce(20)))
            request_token = OAuthRequestToken(**request_token_params)
            session(request_token).flush()
        token_params = dict(
            consumer_token_id=consumer_token._id,
            request_token_id=request_token._id,
            user_id=user._id,
            is_bearer=True)
        token = OAuthAccessToken.query.get(**token_params)
        if token is None:
            token = OAuthAccessToken(**token_params)
            session(token).flush()
        return token.api_key



#       Licensed to the Apache Software Foundation (ASF) under one
#       or more contributor license agreements.  See the NOTICE file
#       distributed with this work for additional information
#       regarding copyright ownership.  The ASF licenses this file
#       to you under the Apache License, Version 2.0 (the
#       "License"); you may not use this file except in compliance
#       with the License.  You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#       Unless required by applicable law or agreed to in writing,
#       software distributed under the License is distributed on an
#       "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#       KIND, either express or implied.  See the License for the
#       specific language governing permissions and limitations
#       under the License.

import json
import logging

from allura.lib.import_api import AlluraImportApiClient


log = logging.getLogger(__name__)

def load_data(doc_file_name=None, optparser=None, options=None):
    import_options = {}
    for s in options.import_opts:
        k, v = s.split('=', 1)
        if v == 'false':
            v = False
        import_options[k] = v

    user_map = {}
    if options.user_map_file:
        f = open(options.user_map_file)
        try:
            user_map = json.load(f)
            if type(user_map) is not type({}):
                raise ValueError
            for k, v in user_map.iteritems():
                print k, v
                if not isinstance(k, basestring) or not isinstance(v, basestring):
                    raise ValueError
        except ValueError:
            optparser.error('--user-map should specify JSON file with format {"original_user": "sf_user", ...}')
        finally:
            f.close()

    import_options['user_map'] = user_map

    cli = AlluraImportApiClient(options.base_url, options.token, options.verbose)
    doc_txt = open(doc_file_name).read()

    if options.wiki:
        import_wiki(cli, options.project, options.wiki, options, doc_txt)


def import_wiki(cli, project, tool, options, doc_txt):
    url = '/rest/{neighborhood}/{project}/{tool}'.format(
            neighborhood=options.neighborhood,
            project=project,
            tool=tool,
            )
    log.info('Loading trac wiki in to %s', url)
    doc = json.loads(doc_txt)
    if 'wiki' in doc and 'default' in doc['wiki'] and 'artifacts' in doc['wiki']['default']:
        pages = doc['trackers']['default']['artifacts']
    else:
        pages = doc
    if options.verbose:
        print "Processing %d pages" % len(pages)
    for page in pages:
        title = page.pop('title').encode('utf-8')
        page['text'] = page['text'].encode('utf-8')
        page['labels'] = page['labels'].encode('utf-8')
        r = cli.call(url + '/' + title, **page)
        assert r == {}
        print 'Imported wiki page %s' % title



#       Licensed to the Apache Software Foundation (ASF) under one
#       or more contributor license agreements.  See the NOTICE file
#       distributed with this work for additional information
#       regarding copyright ownership.  The ASF licenses this file
#       to you under the Apache License, Version 2.0 (the
#       "License"); you may not use this file except in compliance
#       with the License.  You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#       Unless required by applicable law or agreed to in writing,
#       software distributed under the License is distributed on an
#       "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#       KIND, either express or implied.  See the License for the
#       specific language governing permissions and limitations
#       under the License.

import argparse
import logging
from tempfile import NamedTemporaryFile
from tg.decorators import cached_property

from tracwikiimporter.scripts.wiki_from_trac.extractors import WikiExporter
from tracwikiimporter.scripts.wiki_from_trac.loaders import load_data

from allura.scripts import ScriptTask


log = logging.getLogger(__name__)


class WikiFromTrac(ScriptTask):
    """Import Trac Wiki to Allura Wiki"""
    @classmethod
    def parser(cls):
        parser = argparse.ArgumentParser(description='Import wiki from '
            'Trac to allura wiki')

        parser.add_argument('trac_url', type=str, help='Trac URL')
        parser.add_argument('-t', '--token', dest='token',
                            help='OAuth bearer token (generate at /auth/oauth/)')
        parser.add_argument('-p', '--project', dest='project', help='Project to import to')
        parser.add_argument('-n', '--neighborhood', dest='neighborhood', help="URL prefix of destination neighborhood (e.g. 'p')", default='p')
        parser.add_argument('-f', '--forum', dest='forum', help='Forum tool to import to')
        parser.add_argument('-w', '--wiki', dest='wiki', help='Wiki tool to import to')
        parser.add_argument('-u', '--base-url', dest='base_url', default='https://sourceforge.net', help='Base Allura (%(default)s for default)')
        parser.add_argument('-o', dest='import_opts', default=[], action='append', help='Specify import option(s)', metavar='opt=val')
        parser.add_argument('--user-map', dest='user_map_file', help='Map original users to SF.net users', metavar='JSON_FILE')
        parser.add_argument('--validate', dest='validate', action='store_true', help='Validate import data')
        parser.add_argument('-v', '--verbose', dest='verbose', action='store_true', help='Verbose operation')
        parser.add_argument('-c', '--continue', dest='cont', action='store_true', help='Continue import into existing tracker')
        parser.add_argument('-C', '--converter', dest='converter',
                            default='html2text',
                            help='Converter to use on wiki text. '
                                 'Available options: '
                                 'html2text (default) or regex')

        return parser

    @classmethod
    def execute(cls, options):
        options.neighborhood = options.neighborhood.strip('/')
        with NamedTemporaryFile() as f:
            WikiExporter(options.trac_url, options).export(f)
            f.flush()
            load_data(f.name, cls.parser(), options)


if __name__ == '__main__':
    WikiFromTrac.main()
